/************************************************************************
 *   IRC - Internet Relay Chat, src/modules.c
 *   Copyright (C) 1990 Jarkko Oikarinen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: modules.c,v 1.4 2001/09/16 22:45:05 asuffield Exp $
 */

#include "config.h"

#ifndef STATIC_MODULES
#include <dlfcn.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>

#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#include "modules.h"
#include "s_log.h"
#include "ircd.h"
#include "client.h"
#include "send.h"
#include "handlers.h"
#include "numeric.h"
#include "parse.h"
#include "ircd_defs.h"
#include "irc_string.h"
#include "memory.h"
#include "tools.h"
#include "list.h"
#include "umodes.h"
#include "fileio.h"
#include "paths.h"

static char modules_version[] = "1.0.0";

static dlink_list module_list;

#define MODS_INCREMENT 10
int num_mods = 0;

static int mo_modlist(struct Client *, struct Client *, int, char **);

struct Message modlist_msgtab = {
  "MODLIST", 0, 1, 0, MFLG_SLOW, 0,
  {m_unregistered, mo_modlist, m_ignore},
  NULL, NULL
};

struct module_handle *mh_module_system = NULL;

#ifndef STATIC_MODULES

#ifndef RTLD_NOW
#define RTLD_NOW RTLD_LAZY	/* openbsd deficiency */
#endif

static dlink_list mod_paths;

static int mo_modload(struct Client *, struct Client *, int, char **);
static int mo_modreload(struct Client *, struct Client *, int, char **);
static int mo_modunload(struct Client *, struct Client *, int, char **);
static int mo_modrestart(struct Client *, struct Client *, int, char **);

struct Message modload_msgtab = {
  "MODLOAD", 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, mo_modload, m_ignore},
  NULL, NULL
};

struct Message modunload_msgtab = {
  "MODUNLOAD", 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, mo_modunload, m_ignore},
  NULL, NULL
};

struct Message modreload_msgtab = {
  "MODRELOAD", 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, mo_modreload, m_ignore},
  NULL, NULL
};

struct Message modrestart_msgtab = {
  "MODRESTART", 0, 1, 0, MFLG_SLOW, 0,
  {m_unregistered, mo_modrestart, m_ignore},
  NULL, NULL
};

static char *
mod_find_path(char *path)
{
  dlink_node *pathst = mod_paths.head;
  char *p;

  if (!pathst)
    return NULL;

  for (; pathst; pathst = pathst->next)
    {
      p = (char *) pathst->data;

      if (!strcmp(path, p))
	return p;
    }

  return NULL;
}

void
mod_add_path(char *path)
{
  char *p;
  dlink_node *node;
  int len;

  if (mod_find_path(path))
    return;

  len = strlen(path);
  if (path[len - 1] == '/')
    DupString(p, path);
  else
    {
      /* Ensure that paths end with a / */
      p = (char *)MyMalloc(len + 2);
      strncpy_irc(p, path, len);
      p[len] = '/';
      p[len + 1] = '\0';
    }
  node = make_dlink_node();
  dlinkAdd(p, node, &mod_paths);
}

void
mod_clear_paths(void)
{
  dlink_node *node, *next;

  for (node = mod_paths.head; node; node = next)
    {
      next = node->next;
      dlinkDelete(node, &mod_paths);
      MyFree(node->data);
      MyFree(node);
    }
}

char *
irc_basename(char *path)
{
  char *mod_basename, *s;
  int len;

  if (!(s = strrchr(path, '/')))
    s = path;
  else
    s++;

  len = strlen(s);
  /* Remove a trailing .la if it exists */
  if (s[len - 3] == '.' &&
      s[len - 2] == 'l' &&
      s[len - 1] == 'a')
    len -= 3;

  mod_basename = MyMalloc(len + 1);

  strncpy_irc(mod_basename, s, len);

  return mod_basename;
}

/* Like modprobe, this goes hunting for the named module and tries
 *  to load it. Stop at the first file that exists
 */
int
probe_module(char *path)
{
  char modpath[MAXPATHLEN];
  dlink_node *pathst;
  char *mpath;

  struct stat statbuf;

  for (pathst = mod_paths.head; pathst; pathst = pathst->next)
    {
      mpath = (char *) pathst->data;

      snprintf(modpath, MAXPATHLEN, "%s%s.la", mpath, path);
      if (stat(modpath, &statbuf) == 0)
	return load_module(modpath);
    }

  sendto_ops_flag(UMODE_MODULES, "Cannot locate module %s", path);
  log(L_WARN, "Cannot locate module %s", path);
  return -1;
}


/*
 * load_a_module
 *
 * inputs	- path name of module
 * output	- -1 if error 0 if success
 * side effects - loads a module if successful
 */
int
load_module(char *la_path)
{
  FBFILE *in;
  void *tmpptr = NULL;
  char *mod_basename;
  struct module_handle *(*initfunc) (void) = NULL;
  char buf[512];
  char *p, *path = NULL, *modpath;
  struct module_handle *mh;
  char **self_name, **dancer_version, **module_name;
  int old_module_count;

  if ((in = fbopen(la_path, "r")) == 0)
    {
      sendto_ops_flag(UMODE_MODULES, "Cannot open %s", la_path);
      return -1;
    }
  while (fbgets(buf, sizeof(buf), in))
    {
      if (!ircncmp(buf, "dlname=", 7))
	{
	  path = buf + 8;
	  break;
	}
    }
  fbclose(in);
  if (!path)
    {
      sendto_ops_flag(UMODE_MODULES, "%s does not contain a dlname!",
		      la_path);
      log(L_WARN, "%s does not contain a dlname!", la_path);
      return -1;
    }

  if ((p = strchr(path, '\'')))
    *p = '\0';

  /* OK, I have the target based from wherever the .la was. Append it to the original path,
   * minus the name of the .la
   *  -- asuffield
   */
  modpath = MyMalloc(strlen(path) + strlen(la_path));
  strcpy(modpath, la_path);
  if ((p = strrchr(modpath, '/')))
    strcpy(p + 1, path);
  else
    strcpy(modpath, path);

  mod_basename = irc_basename(modpath);

  /* Clip the version number if present */
  if ((p = strrchr(mod_basename, '-')))
    *p = '\0';

  if (findmodule_byname(mod_basename))
    {
      sendto_ops_flag(UMODE_MODULES, "Module %s already loaded",
		      mod_basename);
      log(L_INFO, "Module %s already loaded", mod_basename);
      MyFree(mod_basename);
      MyFree(modpath);
      return -1;
    }

  tmpptr = dlopen(modpath, RTLD_NOW | RTLD_GLOBAL);
  MyFree(modpath);

  if (tmpptr == NULL)
    {
      const char *err = dlerror();

      sendto_ops_flag(UMODE_MODULES,
		      "Error loading module %s: %s", mod_basename, err);
      log(L_WARN, "Error loading module %s: %s", mod_basename, err);
      MyFree(mod_basename);
      return -1;
    }

  dancer_version = (char **) dlsym(tmpptr, "_dancer_version");
  if ((dancer_version == NULL) || (*dancer_version == NULL))
    {
      sendto_ops_flag(UMODE_MODULES, "Module %s has no _dancer_version",
		      mod_basename);
      log(L_WARN, "Module %s has no _dancer_version", mod_basename);
      (void) dlclose(tmpptr);
      MyFree(mod_basename);
      return -1;
    }

  if (irccmp(*dancer_version, VERSION))
    {
      sendto_ops_flag(UMODE_MODULES,
		      "Module %s has _dancer_version == %s, which doesn't match my %s",
		      mod_basename, *dancer_version, VERSION);
      log(L_WARN,
	  "Module %s has _dancer_version == %s, which doesn't match my %s",
	  mod_basename, *dancer_version, VERSION);
      (void) dlclose(tmpptr);
      MyFree(mod_basename);
      return -1;
    }

  module_name = (char **) dlsym(tmpptr, "_name");
  if ((module_name == NULL) || (*module_name == NULL))
    {
      sendto_ops_flag(UMODE_MODULES, "Module %s has no _name",
		      mod_basename);
      log(L_WARN, "Module %s has no _name", mod_basename);
      (void) dlclose(tmpptr);
      MyFree(mod_basename);
      return -1;
    }

  if (irccmp(*module_name, mod_basename))
    {
      sendto_ops_flag(UMODE_MODULES, "Module %s is really %s",
		      mod_basename, *module_name);
      log(L_INFO, "Module %s is really %s",
	  mod_basename, *module_name);
      MyFree(mod_basename);
      DupString(mod_basename, *module_name);
    }

  if (findmodule_byname(mod_basename))
    {
      sendto_ops_flag(UMODE_MODULES, "Module %s already loaded", mod_basename);
      log(L_INFO, "Module %s already loaded", mod_basename);
      MyFree(mod_basename);
      return -1;
    }

  initfunc = (mod_init_func *) dlsym(tmpptr, "_modinit");
  if (initfunc == NULL
      && (initfunc = (mod_init_func *) dlsym(tmpptr, "__modinit")) == NULL)
    {
      sendto_ops_flag(UMODE_MODULES,
		      "Module %s has no _modinit() function", mod_basename);
      log(L_WARN, "Module %s has no _modinit() function", mod_basename);
      (void) dlclose(tmpptr);
      MyFree(mod_basename);
      return -1;
    }

  if (!(mh = initfunc()))
    {
      sendto_ops_flag(UMODE_MODULES,
		      "Module %s failed to return a valid handle, aborting load",
		      mod_basename);
      log(L_WARN, "Module %s failed to return a valid handle, aborting load",
	  mod_basename);
      /* Just in case... */
      if ((mh = findmodule_byname(mod_basename)))
	deregister_module(mh, UNLOAD_ERROR);
      (void) dlclose(tmpptr);
      MyFree(mod_basename);
      return -1;
    }

  mh->code = tmpptr;

  sendto_ops_flag(UMODE_MODULES,
		  "Module %s (%s) version: %s loaded at 0x%p",
		  mh->name, mh->desc, mh->version, tmpptr);
  log(L_WARN, "Module %s (%s) version: %s loaded at 0x%p",
      mh->name, mh->desc, mh->version, tmpptr);
  MyFree(mod_basename);
  return 0;
}

/* load a module .. */
static int
mo_modload(struct Client *client_p, struct Client *source_p, int parc,
	   char **parv)
{
  char *m_bn;

  if (!HasUmode(source_p, UMODE_MODULES))
    {
      sendto_one(source_p, ":%s NOTICE %s :You have no A flag", me.name,
		 parv[0]);
      return 0;
    }

  m_bn = irc_basename(parv[1]);

  if (findmodule_byname(m_bn))
    {
      sendto_one(source_p, ":%s NOTICE %s :Module %s is already loaded",
		 me.name, source_p->name, m_bn);
      return 0;
    }

  probe_module(parv[1]);

  MyFree(m_bn);
  return 0;
}


/* unload a module .. */
static int
mo_modunload(struct Client *client_p, struct Client *source_p, int parc,
	     char **parv)
{
  struct module_handle *mh;
  char *m_bn;

  if (!HasUmode(source_p, UMODE_MODULES))
    {
      sendto_one(source_p, ":%s NOTICE %s :You have no A flag",
		 me.name, parv[0]);
      return 0;
    }

  m_bn = irc_basename(parv[1]);

  if (!(mh = findmodule_byname(m_bn)))
    {
      sendto_one(source_p, ":%s NOTICE %s :Module %s is not loaded",
		 me.name, source_p->name, m_bn);
      MyFree(m_bn);
      return 0;
    }

  if (mh->code)
    unload_module(mh, UNLOAD_USER);
  else
    sendto_one(source_p, ":%s NOTICE %s :Cannot unload module %s",
	       me.name, source_p->name, mh->name);
  
  MyFree(m_bn);

  return 0;
}

static int
mo_modreload(struct Client *client_p, struct Client *source_p, int parc,
	     char **parv)
{
  struct module_handle *mh;
  char *m_bn;

  if (!HasUmode(source_p, UMODE_MODULES))
    {
      sendto_one(source_p, ":%s NOTICE %s :You have no A flag",
		 me.name, parv[0]);
      return 0;
    }

  m_bn = irc_basename(parv[1]);

  if (!(mh = findmodule_byname(m_bn)))
    {
      sendto_one(source_p, ":%s NOTICE %s :Module %s is not loaded",
		 me.name, source_p->name, m_bn);
      MyFree(m_bn);
      return 0;
    }

  if (unload_module(mh, UNLOAD_RELOAD) == -1)
    {
      sendto_one(source_p, ":%s NOTICE %s :Cannot unload module %s",
		 me.name, source_p->name, mh->name);
      MyFree(m_bn);
      return 0;
    }

  probe_module(parv[1]);

  MyFree(m_bn);

  return 0;
}

/* unload and reload all modules */
static int
mo_modrestart(struct Client *client_p, struct Client *source_p, int parc,
	      char **parv)
{
  dlink_node *node, *next;
  struct module_handle *mh;
  int unload_count = 0;

  if (!HasUmode(source_p, UMODE_MODULES))
    {
      sendto_one(source_p, ":%s NOTICE %s :You have no A flag",
		 me.name, parv[0]);
      return 0;
    }

  sendto_one(source_p, ":%s NOTICE %s :Reloading all modules",
	     me.name, parv[0]);


  for (node = module_list.head; node; node = next)
    {
      next = node->next;
      mh = (struct module_handle *) node->data;
      if (mh->code)
	unload_module(mh, UNLOAD_RELOAD);
    }

  clear_module_conf();
  load_module_conf(mod_config_file);

  sendto_ops_flag(UMODE_MODULES,
		  "Module Restart: %d modules unloaded, %d modules loaded",
		  unload_count, num_mods);
  log(L_WARN, "Module Restart: %d modules unloaded, %d modules loaded",
      unload_count, num_mods);

  return 0;
}

void
load_module_conf(char *path)
{
  FBFILE *in;
  char buf[512];
  char *p;

  if ((in = fbopen(path, "r")) == 0)
    {
      sendto_ops_flag(UMODE_SERVNOTICE, "Cannot open %s", path);
      log(L_WARN, "Cannot open %s", path);
      return;
    }
  while (fbgets(buf, sizeof(buf), in))
    {
      if (buf[0] == '#')
	continue;
      /* XXX - Hack, remove trailing \n */
      buf[strlen(buf) - 1] = '\0';
      if (!ircncmp(buf, "path ", 5))
	{
	  for (p = buf + 5; p && (*p == ' '); p++);
	  if (*p)
	    mod_add_path(p);
	}
      if (!ircncmp(buf, "load ", 5))
	{
	  for (p = buf + 5; p && (*p == ' '); p++);
	  if (*p)
	    probe_module(p);
	}
    }
  fbclose(in);
}

void
clear_module_conf()
{
  mod_clear_paths();
}

#endif /* STATIC_MODULES */

int
unload_module(struct module_handle *mh, int type)
{
  if (!mh)
    return -1;

  log(L_INFO, "Unloading module %s (%s)",
      mh->name, mh->desc);
  sendto_ops_flag(UMODE_MODULES, "Unloading module %s (%s)",
		  mh->name, mh->desc);

  deregister_module(mh, type);

  if (mh->code)
#ifdef STATIC_MODULES
    sendto_ops_flag(UMODE_DEBUG, "WTF: mh->code != NULL, but I built with STATIC_MODULES (mh->name == %s)",
		    mh->name);
#else
    dlclose(mh->code);
#endif

  if (num_mods == 0)
    sendto_ops_flag(UMODE_DEBUG, "WTF: num_mods == 0 but I just unloaded a module");
  else
    num_mods--;

  return 0;
}

/* list modules .. */
static int
mo_modlist(struct Client *client_p, struct Client *source_p, int parc,
	   char **parv)
{
  struct module_handle *mh;
  dlink_node *node;

  if (!HasUmode(source_p, UMODE_MODULES))
    {
      sendto_one(source_p, ":%s NOTICE %s :You have no A flag",
		 me.name, parv[0]);
      return 0;
    }

  if (parc > 1)
    {
      sendto_one(source_p,
		 ":%s NOTICE %s :Listing modules matching string '%s'...",
		 me.name, parv[0], parv[1]);
    }
  else
    {
      sendto_one(source_p, ":%s NOTICE %s :Listing all modules...",
		 me.name, parv[0]);
    }

  for (node = module_list.head; node; node = node->next)
    {
      mh = (struct module_handle *)mh;
      if (parc == 1 || match(parv[1], mh->name))
	{
	  sendto_one(source_p, form_str(RPL_MODLIST), me.name, parv[0],
		     mh->name, mh->code,
		     mh->version ? mh->version : "(none)",
		     mh->desc ? mh->desc : "(none)");
	}
    }
  sendto_one(source_p, ":%s NOTICE %s :Done.", me.name, parv[0]);
  return 0;
}

struct module_handle *
register_module(const char *name, const char *desc, const char *version, mod_deinit_func *deinit)
{
  struct module_handle *mh;
  dlink_node *node;

  /* These are _required_ */
  if (!name || !desc || !version)
    return NULL;

  mh = (struct module_handle *) MyMalloc(sizeof(struct module_handle));
  memset(mh, 0, sizeof(struct module_handle));

  DupString(mh->name, name);
  DupString(mh->desc, desc);
  DupString(mh->version, version);
  mh->deinit = deinit;

  node = make_dlink_node();
  dlinkAdd(mh, node, &module_list);

  return mh;
}

void
deregister_module(struct module_handle *mh, int type)
{
  struct message_link *mlink, *next;
  dlink_node *node;

  if (mh->deinit)
    mh->deinit(type);

  MyFree(mh->name);
  MyFree(mh->desc);
  MyFree(mh->version);

  for (mlink = mh->msgs; mlink; mlink = next)
    {
      next = mlink->next;
      mod_del_cmd(mlink->msg);
    }

  for (node = module_list.head; node; node = node->next)
    if (node->data == mh)
      {
	dlinkDelete(node, &module_list);
	MyFree(node);
	break;
      }

  MyFree(mh);
}

void
modules_init(void)
{
  if (mh_module_system)
    {
      log(L_CRIT, "modules_init() called twice, aborting");
      exit(-1);
    }

  mh_module_system =
    register_module("modules", "Module control system", modules_version, NULL);

#ifndef STATIC_MODULES
  mod_add_cmd(&modload_msgtab, mh_module_system);
  mod_add_cmd(&modunload_msgtab, mh_module_system);
  mod_add_cmd(&modreload_msgtab, mh_module_system);
  mod_add_cmd(&modlist_msgtab, mh_module_system);
  mod_add_cmd(&modrestart_msgtab, mh_module_system);
#else
  call_module_inits();
#endif
}

/* Unload all modules and shut down the module system */
void
modules_deinit(void)
{
  dlink_node *node, *next;

  for (node = module_list.head; node; node = next)
    {
      next = node->next;
      unload_module((struct module_handle *)node->data, UNLOAD_SHUTDOWN);
    }

  /* This got whacked by that loop as well */
  mh_module_system = NULL;
}

struct module_handle *
findmodule_byname(char *name)
{
  dlink_node *node;
  struct module_handle *mh;

  for (node = module_list.head; node; node = node->next)
    {
      mh = (struct module_handle *) node->data;
      if (!irccmp(mh->name, name))
	return mh;
    }
  return NULL;
}
